@@ -3,9 +3,11 @@ import inspect
 import os
 import sys
 import typing
+import warnings
 
 from .. import utils, helpers, errors, password as pwd_mod
-from ..tl import types, functions
+from ..tl import types, functions, custom
+from .._updates import SessionState
 
 if typing.TYPE_CHECKING:
     from .telegramclient import TelegramClient
@@ -91,7 +93,7 @@ class AuthMethods:
                 # Starting as a bot account
                 await client.start(bot_token=bot_token)
 
-                # Starting as an user account
+                # Starting as a user account
                 await client.start(phone)
                 # Please enter the code you received: 12345
                 # Please enter your password: *******
@@ -138,7 +140,29 @@ class AuthMethods:
         if not self.is_connected():
             await self.connect()
 
-        if await self.is_user_authorized():
+        # Rather than using `is_user_authorized`, use `get_me`. While this is
+        # more expensive and needs to retrieve more data from the server, it
+        # enables the library to warn users trying to login to a different
+        # account. See #1172.
+        me = await self.get_me()
+        if me is not None:
+            # The warnings here are on a best-effort and may fail.
+            if bot_token:
+                # bot_token's first part has the bot ID, but it may be invalid
+                # so don't try to parse as int (instead cast our ID to string).
+                if bot_token[:bot_token.find(':')] != str(me.id):
+                    warnings.warn(
+                        'the session already had an authorized user so it did '
+                        'not login to the bot account using the provided '
+                        'bot_token (it may not be using the user you expect)'
+                    )
+            elif phone and not callable(phone) and utils.parse_phone(phone) != me.phone:
+                warnings.warn(
+                    'the session already had an authorized user so it did '
+                    'not login to the user account using the provided '
+                    'phone (it may not be using the user you expect)'
+                )
+
             return self
 
         if not bot_token:
@@ -348,7 +372,7 @@ class AuthMethods:
             self._tos = result.terms_of_service
             raise errors.PhoneNumberUnoccupiedError(request=request)
 
-        return self._on_login(result.user)
+        return await self._on_login(result.user)
 
     async def sign_up(
             self: 'TelegramClient',
@@ -404,6 +428,23 @@ class AuthMethods:
         if me:
             return me
 
+        # To prevent abuse, one has to try to sign in before signing up. This
+        # is the current way in which Telegram validates the code to sign up.
+        #
+        # `sign_in` will set `_tos`, so if it's set we don't need to call it
+        # because the user already tried to sign in.
+        #
+        # We're emulating pre-layer 104 behaviour so except the right error:
+        if not self._tos:
+            try:
+                return await self.sign_in(
+                    phone=phone,
+                    code=code,
+                    phone_code_hash=phone_code_hash,
+                )
+            except errors.PhoneNumberUnoccupiedError:
+                pass  # code is correct and was used, now need to sign in
+
         if self._tos and self._tos.text:
             if self.parse_mode:
                 t = self.parse_mode.unparse(self._tos.text, self._tos.entities)
@@ -426,9 +467,9 @@ class AuthMethods:
             await self(
                 functions.help.AcceptTermsOfServiceRequest(self._tos.id))
 
-        return self._on_login(result.user)
+        return await self._on_login(result.user)
 
-    def _on_login(self, user):
+    async def _on_login(self, user):
         """
         Callback called whenever the login or sign up process completes.
 
@@ -438,13 +479,17 @@ class AuthMethods:
         self._self_input_peer = utils.get_input_peer(user, allow_self=False)
         self._authorized = True
 
+        state = await self(functions.updates.GetStateRequest())
+        self._message_box.load(SessionState(0, 0, 0, state.pts, state.qts, int(state.date.timestamp()), state.seq, 0), [])
+
         return user
 
     async def send_code_request(
             self: 'TelegramClient',
             phone: str,
             *,
-            force_sms: bool = False) -> 'types.auth.SentCode':
+            force_sms: bool = False,
+            _retry_count: int = 0) -> 'types.auth.SentCode':
         """
         Sends the Telegram code needed to login to the given phone number.
 
@@ -474,7 +519,14 @@ class AuthMethods:
                 result = await self(functions.auth.SendCodeRequest(
                     phone, self.api_id, self.api_hash, types.CodeSettings()))
             except errors.AuthRestartError:
-                return await self.send_code_request(phone, force_sms=force_sms)
+                if _retry_count > 2:
+                    raise
+                return await self.send_code_request(
+                    phone, force_sms=force_sms, _retry_count=_retry_count+1)
+
+            # TODO figure out when/if/how this can happen
+            if isinstance(result, types.auth.SentCodeSuccess):
+                raise RuntimeError('logged in right after sending the code')
 
             # If we already sent a SMS, do not resend the code (hash may be empty)
             if isinstance(result.type, types.auth.SentCodeTypeSms):
@@ -489,17 +541,72 @@ class AuthMethods:
         self._phone = phone
 
         if force_sms:
-            result = await self(
-                functions.auth.ResendCodeRequest(phone, phone_hash))
+            try:
+                result = await self(
+                    functions.auth.ResendCodeRequest(phone, phone_hash))
+            except errors.PhoneCodeExpiredError:
+                if _retry_count > 2:
+                    raise
+                self._phone_code_hash.pop(phone, None)
+                self._log[__name__].info(
+                    "Phone code expired in ResendCodeRequest, requesting a new code"
+                )
+                return await self.send_code_request(
+                    phone, force_sms=False, _retry_count=_retry_count+1)
+
+            if isinstance(result, types.auth.SentCodeSuccess):
+                raise RuntimeError('logged in right after resending the code')
 
             self._phone_code_hash[phone] = result.phone_code_hash
 
         return result
 
+    async def qr_login(self: 'TelegramClient', ignored_ids: typing.List[int] = None) -> custom.QRLogin:
+        """
+        Initiates the QR login procedure.
+
+        Note that you must be connected before invoking this, as with any
+        other request.
+
+        It is up to the caller to decide how to present the code to the user,
+        whether it's the URL, using the token bytes directly, or generating
+        a QR code and displaying it by other means.
+
+        See the documentation for `QRLogin` to see how to proceed after this.
+
+        Arguments
+            ignored_ids (List[`int`]):
+                List of already logged-in user IDs, to prevent logging in
+                twice with the same user.
+
+        Returns
+            An instance of `QRLogin`.
+
+        Example
+            .. code-block:: python
+
+                def display_url_as_qr(url):
+                    pass  # do whatever to show url as a qr to the user
+
+                qr_login = await client.qr_login()
+                display_url_as_qr(qr_login.url)
+
+                # Important! You need to wait for the login to complete!
+                await qr_login.wait()
+
+                # If you have 2FA enabled, `wait` will raise `telethon.errors.SessionPasswordNeededError`.
+                # You should except that error and call `sign_in` with the password if this happens.
+        """
+        qr_login = custom.QRLogin(self, ignored_ids or [])
+        await qr_login.recreate()
+        return qr_login
+
     async def log_out(self: 'TelegramClient') -> bool:
         """
         Logs out Telegram and deletes the current ``*.session`` file.
 
+        The client is unusable after logging out and a new instance should be created.
+
         Returns
             `True` if the operation was successful.
 
@@ -517,10 +624,10 @@ class AuthMethods:
         self._bot = None
         self._self_input_peer = None
         self._authorized = False
-        self._state_cache.reset()
 
         await self.disconnect()
         self.session.delete()
+        self.session = None
         return True
 
     async def edit_2fa(
