@@ -4,6 +4,7 @@ import os
 import pathlib
 import typing
 import inspect
+import asyncio
 
 from ..crypto import AES
 
@@ -23,6 +24,8 @@ if typing.TYPE_CHECKING:
 MIN_CHUNK_SIZE = 4096
 MAX_CHUNK_SIZE = 512 * 1024
 
+# 2021-01-15, users reported that `errors.TimeoutError` can occur while downloading files.
+TIMED_OUT_SLEEP = 1
 
 class _DirectDownloadIter(RequestIter):
     async def _init(
@@ -36,6 +39,7 @@ class _DirectDownloadIter(RequestIter):
         self._chunk_size = chunk_size
         self._last_part = None
         self._msg_data = msg_data
+        self._timed_out = False
 
         self._exported = dc_id and self.client.session.dc_id != dc_id
         if not self._exported:
@@ -70,11 +74,22 @@ class _DirectDownloadIter(RequestIter):
     async def _request(self):
         try:
             result = await self.client._call(self._sender, self.request)
+            self._timed_out = False
             if isinstance(result, types.upload.FileCdnRedirect):
                 raise NotImplementedError  # TODO Implement
             else:
                 return result.bytes
 
+        except errors.TimeoutError as e:
+            if self._timed_out:
+                self.client._log[__name__].warning('Got two timeouts in a row while downloading file')
+                raise
+
+            self._timed_out = True
+            self.client._log[__name__].info('Got timeout while downloading file, retrying once')
+            await asyncio.sleep(TIMED_OUT_SLEEP)
+            return await self._request()
+
         except errors.FileMigrateError as e:
             self.client._log[__name__].info('File lives in another DC')
             self._sender = await self.client._borrow_exported_sender(e.new_dc)
@@ -127,12 +142,12 @@ class _DirectDownloadIter(RequestIter):
 
 
 class _GenericDownloadIter(_DirectDownloadIter):
-    async def _load_next_chunk(self, mask=MIN_CHUNK_SIZE - 1):
+    async def _load_next_chunk(self):
         # 1. Fetch enough for one chunk
         data = b''
 
         # 1.1. ``bad`` is how much into the data we have we need to offset
-        bad = self.request.offset & mask
+        bad = self.request.offset % self.request.limit
         before = self.request.offset
 
         # 1.2. We have to fetch from a valid offset, so remove that bad part
@@ -256,11 +271,11 @@ class DownloadMethods:
 
         if isinstance(photo, (types.UserProfilePhoto, types.ChatPhoto)):
             dc_id = photo.dc_id
-            which = photo.photo_big if download_big else photo.photo_small
             loc = types.InputPeerPhotoFileLocation(
-                peer=await self.get_input_entity(entity),
-                local_id=which.local_id,
-                volume_id=which.volume_id,
+                # min users can be used to download profile photos
+                # self.get_input_entity would otherwise not accept those
+                peer=utils.get_input_peer(entity, check_hash=False),
+                photo_id=photo.photo_id,
                 big=download_big
             )
         else:
@@ -385,6 +400,11 @@ class DownloadMethods:
         if isinstance(media, str):
             media = utils.resolve_bot_file_id(media)
 
+        if isinstance(media, types.MessageService):
+            if isinstance(message.action,
+                          types.MessageActionChatEditPhoto):
+                media = media.photo
+
         if isinstance(media, types.MessageMediaWebPage):
             if isinstance(media.webpage, types.WebPage):
                 media = media.webpage.document or media.webpage.photo
@@ -687,7 +707,8 @@ class DownloadMethods:
 
         if chunk_size == request_size \
                 and offset % MIN_CHUNK_SIZE == 0 \
-                and stride % MIN_CHUNK_SIZE == 0:
+                and stride % MIN_CHUNK_SIZE == 0 \
+                and (limit is None or offset % limit == 0):
             cls = _DirectDownloadIter
             self._log[__name__].info('Starting direct file download in chunks of '
                                      '%d at %d, stride %d', request_size, offset, stride)
@@ -725,6 +746,8 @@ class DownloadMethods:
                 return 1, len(thumb.bytes)
             if isinstance(thumb, types.PhotoSize):
                 return 1, thumb.size
+            if isinstance(thumb, types.PhotoSizeProgressive):
+                return 1, max(thumb.sizes)
             if isinstance(thumb, types.VideoSize):
                 return 2, thumb.size
 
@@ -733,6 +756,13 @@ class DownloadMethods:
 
         thumbs = list(sorted(thumbs, key=sort_thumbs))
 
+        for i in reversed(range(len(thumbs))):
+            # :tl:`PhotoPathSize` is used for animated stickers preview, and the thumb is actually
+            # a SVG path of the outline. Users expect thumbnails to be JPEG files, so pretend this
+            # thumb size doesn't actually exist (#1655).
+            if isinstance(thumbs[i], types.PhotoPathSize):
+                thumbs.pop(i)
+
         if thumb is None:
             return thumbs[-1]
         elif isinstance(thumb, int):
@@ -788,6 +818,11 @@ class DownloadMethods:
         if isinstance(size, (types.PhotoCachedSize, types.PhotoStrippedSize)):
             return self._download_cached_photo_size(size, file)
 
+        if isinstance(size, types.PhotoSizeProgressive):
+            file_size = max(size.sizes)
+        else:
+            file_size = size.size
+
         result = await self.download_file(
             types.InputPhotoFileLocation(
                 id=photo.id,
@@ -796,7 +831,7 @@ class DownloadMethods:
                 thumb_size=size.type
             ),
             file,
-            file_size=size.size,
+            file_size=file_size,
             progress_callback=progress_callback
         )
         return result if file is bytes else file
@@ -883,22 +918,19 @@ class DownloadMethods:
             'END:VCARD\n'
         ).format(f=first_name, l=last_name, p=phone_number).encode('utf-8')
 
+        file = cls._get_proper_filename(
+            file, 'contact', '.vcard',
+            possible_names=[first_name, phone_number, last_name]
+        )
         if file is bytes:
             return result
-        elif isinstance(file, str):
-            file = cls._get_proper_filename(
-                file, 'contact', '.vcard',
-                possible_names=[first_name, phone_number, last_name]
-            )
-            f = open(file, 'wb')
-        else:
-            f = file
+        f = file if hasattr(file, 'write') else open(file, 'wb')
 
         try:
             f.write(result)
         finally:
             # Only close the stream if we opened it
-            if isinstance(file, str):
+            if f != file:
                 f.close()
 
         return file
@@ -915,21 +947,20 @@ class DownloadMethods:
             )
 
         # TODO Better way to get opened handles of files and auto-close
-        in_memory = file is bytes
-        if in_memory:
+        kind, possible_names = self._get_kind_and_names(web.attributes)
+        file = self._get_proper_filename(
+            file, kind, utils.get_extension(web),
+            possible_names=possible_names
+        )
+        if file is bytes:
             f = io.BytesIO()
-        elif isinstance(file, str):
-            kind, possible_names = cls._get_kind_and_names(web.attributes)
-            file = cls._get_proper_filename(
-                file, kind, utils.get_extension(web),
-                possible_names=possible_names
-            )
-            f = open(file, 'wb')
-        else:
+        elif hasattr(file, 'write'):
             f = file
+        else:
+            f = open(file, 'wb')
 
         try:
-            with aiohttp.ClientSession() as session:
+            async with aiohttp.ClientSession() as session:
                 # TODO Use progress_callback; get content length from response
                 # https://github.com/telegramdesktop/tdesktop/blob/c7e773dd9aeba94e2be48c032edc9a78bb50234e/Telegram/SourceFiles/ui/images.cpp#L1318-L1319
                 async with session.get(web.url) as response:
@@ -939,10 +970,10 @@ class DownloadMethods:
                             break
                         f.write(chunk)
         finally:
-            if isinstance(file, str) or file is bytes:
+            if f != file:
                 f.close()
 
-        return f.getvalue() if in_memory else file
+        return f.getvalue() if file is bytes else file
 
     @staticmethod
     def _get_proper_filename(file, kind, extension,
