@@ -19,7 +19,8 @@ class _MessagesIter(RequestIter):
     """
     async def _init(
             self, entity, offset_id, min_id, max_id,
-            from_user, offset_date, add_offset, filter, search, reply_to
+            from_user, offset_date, add_offset, filter, search, reply_to,
+            scheduled
     ):
         # Note that entity being `None` will perform a global search.
         if entity:
@@ -70,6 +71,8 @@ class _MessagesIter(RequestIter):
 
         if filter is None:
             filter = types.InputMessagesFilterEmpty()
+        else:
+            filter = filter() if isinstance(filter, type) else filter
 
         if not self.entity:
             self.request = functions.messages.SearchGlobalRequest(
@@ -77,11 +80,16 @@ class _MessagesIter(RequestIter):
                 filter=filter,
                 min_date=None,
                 max_date=offset_date,
-                offset_rate=None,
+                offset_rate=0,
                 offset_peer=types.InputPeerEmpty(),
                 offset_id=offset_id,
                 limit=1
             )
+        elif scheduled:
+            self.request = functions.messages.GetScheduledHistoryRequest(
+                peer=entity,
+                hash=0
+            )
         elif reply_to is not None:
             self.request = functions.messages.GetRepliesRequest(
                 peer=self.entity,
@@ -94,7 +102,7 @@ class _MessagesIter(RequestIter):
                 min_id=0,
                 hash=0
             )
-        elif search is not None or filter or from_user:
+        elif search is not None or not isinstance(filter, types.InputMessagesFilterEmpty) or from_user:
             # Telegram completely ignores `from_id` in private chats
             ty = helpers._entity_type(self.entity)
             if ty == helpers._EntityType.USER:
@@ -109,7 +117,7 @@ class _MessagesIter(RequestIter):
             self.request = functions.messages.SearchRequest(
                 peer=self.entity,
                 q=search or '',
-                filter=filter() if isinstance(filter, type) else filter,
+                filter=filter,
                 min_date=None,
                 max_date=offset_date,
                 offset_id=offset_id,
@@ -196,7 +204,24 @@ class _MessagesIter(RequestIter):
             message._finish_init(self.client, entities, self.entity)
             self.buffer.append(message)
 
-        if len(r.messages) < self.request.limit:
+        # Not a slice (using offset would return the same, with e.g. SearchGlobal).
+        if isinstance(r, types.messages.Messages):
+            return True
+
+        # Some channels are "buggy" and may return less messages than
+        # requested (apparently, the messages excluded are, for example,
+        # "not displayable due to local laws").
+        #
+        # This means it's not safe to rely on `len(r.messages) < req.limit` as
+        # the stop condition. Unfortunately more requests must be made.
+        #
+        # However we can still check if the highest ID is equal to or lower
+        # than the limit, in which case there won't be any more messages
+        # because the lowest message ID is 1.
+        #
+        # We also assume the API will always return, at least, one message if
+        # there is more to fetch.
+        if not r.messages or r.messages[0].id <= self.request.limit:
             return True
 
         # Get the last message that's not empty (in some rare cases
@@ -253,7 +278,7 @@ class _MessagesIter(RequestIter):
             else:
                 self.request.offset_peer = types.InputPeerEmpty()
 
-            self.request.offset_rate = getattr(response, 'next_rate', None)
+            self.request.offset_rate = getattr(response, 'next_rate', 0)
 
 
 class _IDsIter(RequestIter):
@@ -334,7 +359,8 @@ class MessageMethods:
             wait_time: float = None,
             ids: 'typing.Union[int, typing.Sequence[int]]' = None,
             reverse: bool = False,
-            reply_to: int = None
+            reply_to: int = None,
+            scheduled: bool = False
     ) -> 'typing.Union[_MessagesIter, _IDsIter]':
         """
         Iterator over the messages for the given chat.
@@ -461,6 +487,10 @@ class MessageMethods:
                     a message and replies to it itself, that reply will not
                     be included in the results.
 
+            scheduled (`bool`, optional):
+                If set to `True`, messages which are scheduled will be returned.
+                All other parameter will be ignored for this, except `entity`.
+
         Yields
             Instances of `Message <telethon.tl.custom.message.Message>`.
 
@@ -519,7 +549,8 @@ class MessageMethods:
             add_offset=add_offset,
             filter=filter,
             search=search,
-            reply_to=reply_to
+            reply_to=reply_to,
+            scheduled=scheduled
         )
 
     async def get_messages(self: 'TelegramClient', *args, **kwargs) -> 'hints.TotalList':
@@ -577,21 +608,39 @@ class MessageMethods:
 
     # region Message sending/editing/deleting
 
+    async def _get_comment_data(
+            self: 'TelegramClient',
+            entity: 'hints.EntityLike',
+            message: 'typing.Union[int, types.Message]'
+    ):
+        r = await self(functions.messages.GetDiscussionMessageRequest(
+            peer=entity,
+            msg_id=utils.get_message_id(message)
+        ))
+        m = r.messages[0]
+        chat = next(c for c in r.chats if c.id == m.peer_id.channel_id)
+        return utils.get_input_peer(chat), m.id
+
     async def send_message(
             self: 'TelegramClient',
             entity: 'hints.EntityLike',
             message: 'hints.MessageLike' = '',
             *,
             reply_to: 'typing.Union[int, types.Message]' = None,
+            attributes: 'typing.Sequence[types.TypeDocumentAttribute]' = None,
             parse_mode: typing.Optional[str] = (),
             formatting_entities: typing.Optional[typing.List[types.TypeMessageEntity]] = None,
             link_preview: bool = True,
             file: 'typing.Union[hints.FileLike, typing.Sequence[hints.FileLike]]' = None,
+            thumb: 'hints.FileLike' = None,
             force_document: bool = False,
             clear_draft: bool = False,
-            buttons: 'hints.MarkupLike' = None,
+            buttons: typing.Optional['hints.MarkupLike'] = None,
             silent: bool = None,
-            schedule: 'hints.DateLike' = None
+            background: bool = None,
+            supports_streaming: bool = False,
+            schedule: 'hints.DateLike' = None,
+            comment_to: 'typing.Union[int, types.Message]' = None
     ) -> 'types.Message':
         """
         Sends a message to the specified user, chat or channel.
@@ -626,6 +675,10 @@ class MessageMethods:
                 Whether to reply to a message or not. If an integer is provided,
                 it should be the ID of the message that it should reply to.
 
+            attributes (`list`, optional):
+                Optional attributes that override the inferred ones, like
+                :tl:`DocumentAttributeFilename` and so on.
+
             parse_mode (`object`, optional):
                 See the `TelegramClient.parse_mode
                 <telethon.client.messageparse.MessageParseMethods.parse_mode>`
@@ -642,6 +695,17 @@ class MessageMethods:
                 Sends a message with a file attached (e.g. a photo,
                 video, audio or document). The ``message`` may be empty.
 
+            thumb (`str` | `bytes` | `file`, optional):
+                Optional JPEG thumbnail (for documents). **Telegram will
+                ignore this parameter** unless you pass a ``.jpg`` file!
+                The file must also be small in dimensions and in disk size.
+                Successful thumbnails were files below 20kB and 320x320px.
+                Width/height and dimensions/size ratios may be important.
+                For Telegram to accept a thumbnail, you must provide the
+                dimensions of the underlying media through ``attributes=``
+                with :tl:`DocumentAttributesVideo` or by installing the
+                optional ``hachoir`` dependency.
+
             force_document (`bool`, optional):
                 Whether to send the given file as a document or not.
 
@@ -667,11 +731,29 @@ class MessageMethods:
                 channel or not. Defaults to `False`, which means it will
                 notify them. Set it to `True` to alter this behaviour.
 
+            background (`bool`, optional):
+                Whether the message should be send in background.
+
+            supports_streaming (`bool`, optional):
+                Whether the sent video supports streaming or not. Note that
+                Telegram only recognizes as streamable some formats like MP4,
+                and others like AVI or MKV will not work. You should convert
+                these to MP4 before sending if you want them to be streamable.
+                Unsupported formats will result in ``VideoContentTypeError``.
+
             schedule (`hints.DateLike`, optional):
                 If set, the message won't send immediately, and instead
                 it will be scheduled to be automatically sent at a later
                 time.
 
+            comment_to (`int` | `Message <telethon.tl.custom.message.Message>`, optional):
+                Similar to ``reply_to``, but replies in the linked group of a
+                broadcast channel instead (effectively leaving a "comment to"
+                the specified message).
+
+                This parameter takes precedence over ``reply_to``. If there is
+                no linked chat, `telethon.errors.sgIdInvalidError` is raised.
+
         Returns
             The sent `custom.Message <telethon.tl.custom.message.Message>`.
 
@@ -734,12 +816,18 @@ class MessageMethods:
         if file is not None:
             return await self.send_file(
                 entity, file, caption=message, reply_to=reply_to,
-                parse_mode=parse_mode, force_document=force_document,
+                attributes=attributes, parse_mode=parse_mode,
+                force_document=force_document, thumb=thumb,
                 buttons=buttons, clear_draft=clear_draft, silent=silent,
-                schedule=schedule, formatting_entities=formatting_entities
+                schedule=schedule, supports_streaming=supports_streaming,
+                formatting_entities=formatting_entities,
+                comment_to=comment_to, background=background
             )
 
         entity = await self.get_input_entity(entity)
+        if comment_to is not None:
+            entity, reply_to = await self._get_comment_data(entity, comment_to)
+
         if isinstance(message, types.Message):
             if buttons is None:
                 markup = message.reply_markup
@@ -756,9 +844,11 @@ class MessageMethods:
                     message.media,
                     caption=message.message,
                     silent=silent,
+                    background=background,
                     reply_to=reply_to,
                     buttons=markup,
                     formatting_entities=message.entities,
+                    parse_mode=None,  # explicitly disable parse_mode to force using even empty formatting_entities
                     schedule=schedule
                 )
 
@@ -766,6 +856,7 @@ class MessageMethods:
                 peer=entity,
                 message=message.message or '',
                 silent=silent,
+                background=background,
                 reply_to_msg_id=utils.get_message_id(reply_to),
                 reply_markup=markup,
                 entities=message.entities,
@@ -791,6 +882,7 @@ class MessageMethods:
                 reply_to_msg_id=utils.get_message_id(reply_to),
                 clear_draft=clear_draft,
                 silent=silent,
+                background=background,
                 reply_markup=self.build_reply_markup(buttons),
                 schedule_date=schedule
             )
@@ -805,7 +897,9 @@ class MessageMethods:
                 out=result.out,
                 media=result.media,
                 entities=result.entities,
-                reply_markup=request.reply_markup
+                reply_markup=request.reply_markup,
+                ttl_period=result.ttl_period,
+                reply_to=types.MessageReplyHeader(request.reply_to_msg_id)
             )
             message._finish_init(self, {}, entity)
             return message
@@ -818,6 +912,8 @@ class MessageMethods:
             messages: 'typing.Union[hints.MessageIDLike, typing.Sequence[hints.MessageIDLike]]',
             from_peer: 'hints.EntityLike' = None,
             *,
+            background: bool = None,
+            with_my_score: bool = None,
             silent: bool = None,
             as_album: bool = None,
             schedule: 'hints.DateLike' = None
@@ -850,6 +946,12 @@ class MessageMethods:
                 the person has the chat muted). Set it to `True` to alter
                 this behaviour.
 
+            background (`bool`, optional):
+                Whether the message should be forwarded in background.
+
+            with_my_score (`bool`, optional):
+                Whether forwarded should contain your game score.
+
             as_album (`bool`, optional):
                 This flag no longer has any effect.
 
@@ -924,6 +1026,8 @@ class MessageMethods:
                 id=chunk,
                 to_peer=entity,
                 silent=silent,
+                background=background,
+                with_my_score=with_my_score,
                 schedule_date=schedule
             )
             result = await self(req)
@@ -938,11 +1042,14 @@ class MessageMethods:
             text: str = None,
             *,
             parse_mode: str = (),
+            attributes: 'typing.Sequence[types.TypeDocumentAttribute]' = None,
             formatting_entities: typing.Optional[typing.List[types.TypeMessageEntity]] = None,
             link_preview: bool = True,
             file: 'hints.FileLike' = None,
+            thumb: 'hints.FileLike' = None,
             force_document: bool = False,
-            buttons: 'hints.MarkupLike' = None,
+            buttons: typing.Optional['hints.MarkupLike'] = None,
+            supports_streaming: bool = False,
             schedule: 'hints.DateLike' = None
     ) -> 'types.Message':
         """
@@ -978,6 +1085,10 @@ class MessageMethods:
                 property for allowed values. Markdown parsing will be used by
                 default.
 
+            attributes (`list`, optional):
+                Optional attributes that override the inferred ones, like
+                :tl:`DocumentAttributeFilename` and so on.
+
             formatting_entities (`list`, optional):
                 A list of message formatting entities. When provided, the ``parse_mode`` is ignored.
 
@@ -988,6 +1099,17 @@ class MessageMethods:
                 The file object that should replace the existing media
                 in the message.
 
+            thumb (`str` | `bytes` | `file`, optional):
+                Optional JPEG thumbnail (for documents). **Telegram will
+                ignore this parameter** unless you pass a ``.jpg`` file!
+                The file must also be small in dimensions and in disk size.
+                Successful thumbnails were files below 20kB and 320x320px.
+                Width/height and dimensions/size ratios may be important.
+                For Telegram to accept a thumbnail, you must provide the
+                dimensions of the underlying media through ``attributes=``
+                with :tl:`DocumentAttributesVideo` or by installing the
+                optional ``hachoir`` dependency.
+
             force_document (`bool`, optional):
                 Whether to send the given file as a document or not.
 
@@ -997,6 +1119,13 @@ class MessageMethods:
                 you have signed in as a bot. You can also pass your own
                 :tl:`ReplyMarkup` here.
 
+            supports_streaming (`bool`, optional):
+                Whether the sent video supports streaming or not. Note that
+                Telegram only recognizes as streamable some formats like MP4,
+                and others like AVI or MKV will not work. You should convert
+                these to MP4 before sending if you want them to be streamable.
+                Unsupported formats will result in ``VideoContentTypeError``.
+
             schedule (`hints.DateLike`, optional):
                 If set, the message won't be edited immediately, and instead
                 it will be scheduled to be automatically edited at a later
@@ -1034,7 +1163,7 @@ class MessageMethods:
                 await client.edit_message(message, 'hello!!!')
         """
         if isinstance(entity, types.InputBotInlineMessageID):
-            text = message
+            text = text or message
             message = entity
         elif isinstance(entity, types.Message):
             text = message  # Shift the parameters to the right
@@ -1044,6 +1173,9 @@ class MessageMethods:
         if formatting_entities is None:
             text, formatting_entities = await self._parse_message_text(text, parse_mode)
         file_handle, media, image = await self._file_to_media(file,
+                supports_streaming=supports_streaming,
+                thumb=thumb,
+                attributes=attributes,
                 force_document=force_document)
 
         if isinstance(entity, types.InputBotInlineMessageID):
@@ -1164,7 +1296,8 @@ class MessageMethods:
             message: 'typing.Union[hints.MessageIDLike, typing.Sequence[hints.MessageIDLike]]' = None,
             *,
             max_id: int = None,
-            clear_mentions: bool = False) -> bool:
+            clear_mentions: bool = False,
+            clear_reactions: bool = False) -> bool:
         """
         Marks messages as read and optionally clears mentions.
 
@@ -1198,6 +1331,13 @@ class MessageMethods:
                 If no message is provided, this will be the only action
                 taken.
 
+            clear_reactions (`bool`):
+                Whether the reactions badge should be cleared (so that
+                there are no more reaction notifications) or not for the given entity.
+
+                If no message is provided, this will be the only action
+                taken.
+
         Example
             .. code-block:: python
 
@@ -1220,6 +1360,10 @@ class MessageMethods:
         entity = await self.get_input_entity(entity)
         if clear_mentions:
             await self(functions.messages.ReadMentionsRequest(entity))
+            if max_id is None and not clear_reactions:
+                return True
+        if clear_reactions:
+            await self(functions.messages.ReadReactionsRequest(entity))
             if max_id is None:
                 return True
 
@@ -1238,7 +1382,8 @@ class MessageMethods:
             entity: 'hints.EntityLike',
             message: 'typing.Optional[hints.MessageIDLike]',
             *,
-            notify: bool = False
+            notify: bool = False,
+            pm_oneside: bool = False
     ):
         """
         Pins a message in a chat.
@@ -1259,6 +1404,11 @@ class MessageMethods:
             notify (`bool`, optional):
                 Whether the pin should notify people or not.
 
+            pm_oneside (`bool`, optional):
+                Whether the message should be pinned for everyone or not.
+                By default it has the opposite behaviour of official clients,
+                and it will pin the message for both sides, in private chats.
+
         Example
             .. code-block:: python
 
@@ -1266,7 +1416,7 @@ class MessageMethods:
                 message = await client.send_message(chat, 'Pinotifying is fun!')
                 await client.pin_message(chat, message, notify=True)
         """
-        return await self._pin(entity, message, unpin=False, notify=notify)
+        return await self._pin(entity, message, unpin=False, notify=notify, pm_oneside=pm_oneside)
 
     async def unpin_message(
             self: 'TelegramClient',
@@ -1298,7 +1448,7 @@ class MessageMethods:
         """
         return await self._pin(entity, message, unpin=True, notify=notify)
 
-    async def _pin(self, entity, message, *, unpin, notify=False):
+    async def _pin(self, entity, message, *, unpin, notify=False, pm_oneside=False):
         message = utils.get_message_id(message) or 0
         entity = await self.get_input_entity(entity)
         if message <= 0:  # old behaviour accepted negative IDs to unpin
@@ -1310,15 +1460,15 @@ class MessageMethods:
             id=message,
             silent=not notify,
             unpin=unpin,
+            pm_oneside=pm_oneside
         )
         result = await self(request)
 
-        # Unpinning does not produce a service message
-        if unpin:
-            return
-
-        # Pinning in User chats (just with yourself really) does not produce a service message
-        if helpers._entity_type(entity) == helpers._EntityType.USER:
+        # Unpinning does not produce a service message.
+        # Pinning a message that was already pinned also produces no service message.
+        # Pinning a message in your own chat does not produce a service message,
+        # but pinning on a private conversation with someone else does.
+        if unpin or not result.updates:
             return
 
         # Pinning a message that doesn't exist would RPC-error earlier
